#include "CAalgos.h"
#include "WMutils.h"

#include <math.h>
#include <iostream>
#include <set>
#include <map>
#include <algorithm>
#include <numeric>
#include <functional>
#include <cassert>
#include <fstream>
#include <sstream>
#include <iterator>

#define CHANNEL(l)              \
l.second->radio->channel

//#define DEBUG
#define VERBOSE


class MVCRAclass;


namespace MVCRA_NS
{
  /** ------------------------------------------------------- **/
  /*
  * Abstract base class to assign a cost to the replacement of channel c_to_erase with channel c on node u
  */
  class ChannelCostFunc
  {
  public:
    virtual double GetCost (const pair<unsigned int,set<TLinkItrs> >& p) = 0;

    void SetNodeChannel (int _u, unsigned int _c_selected);
    
    void SetMVCRA (MVCRA* _mvcra) { mvcra = _mvcra; }
    
  protected:
    const vector<vector<int> >& getChangesPerChannel ();
    const set<TLinkItrs>& getQ ();
    // check if a link is repairable
    bool IsRepairable (const TLinkItrs& l);

    int u;
    unsigned c_selected;
    
  private:
    MVCRA* mvcra;
  };

  // The cost is the number of times the channel to be replaced has been set on u (as per changesPerChannel)
  // normalized to the number of channel changes on u
  class ChannelSetCount : public virtual ChannelCostFunc
  {
  public:
    virtual double GetCost (const pair<unsigned int,set<TLinkItrs> >& p);
  };

  // The cost is the number of disrupted links (repairable links are not counted)
  class DisruptedLinkCount : public virtual ChannelCostFunc
  {
  public:
    virtual double GetCost (const pair<unsigned int,set<TLinkItrs> >& p);
  };

  // The cost is the number of extracted links that would be disrupted (repairable links are not counted)
  class DisruptExtrLinkCount : public virtual ChannelCostFunc
  {
  public:
    virtual double GetCost (const pair<unsigned int,set<TLinkItrs> >& p);
  };

  // The cost is the sum of the F/C ratios of the non-repairable links
  class DisruptedLinkFC : public virtual ChannelCostFunc
  {
  public:
    virtual double GetCost (const pair<unsigned int,set<TLinkItrs> >& p);
  };

  class MixedCost : public ChannelSetCount, public DisruptedLinkFC
  {
  public:
    virtual double GetCost (const pair<unsigned int,set<TLinkItrs> >& p);
  };


  /** ------------------------------------------------------- **/

  /*
  * Abstract base class to assign a priority to a link
  */
  class LinkPriority
  {
  public:
    LinkPriority (Topology* _T) : T(_T) {}
    
    virtual double calc (const TLinkItrs& l) = 0;

  protected:
    Topology* T;
  };

  // The priority is the Utot included in the Utot field
  class UtotPriority : public LinkPriority
  {
  public:
    UtotPriority (Topology* _T) : LinkPriority(_T) {}
    
    virtual double calc (const TLinkItrs& l) { return l.second->UTot(); }
  };

  // The priority is the F/C ratio of the link
  class FCPriority : public LinkPriority
  {
  public:
    FCPriority (Topology* _T) : LinkPriority(_T) {}
    
    virtual double calc (const TLinkItrs& l) { return l.second->Flow() / l.second->Rate(); }
  };

  // The priority of a link is its f/c ratio multiplied by the number of links
  // with Utot >= u0 having that link in their collision domain
  class SumUtotPriority : public LinkPriority
  {
  public:
    SumUtotPriority(Topology* _T, double _u0) : LinkPriority(_T), u0(_u0) {}
    
    virtual double calc (const TLinkItrs& l);
    
  private:
    double u0;
  };

  /*
  * Binary predicate used to compare the priorities of links
  */
  class LinkPriorityPred {
      LinkPriority* lp;
    public:
      // constructor
      LinkPriorityPred (LinkPriority* _lp) : lp(_lp) {}
      // operator()
      bool operator()(const TLinkItrs& p, const TLinkItrs& q) {
	return lp->calc(p) < lp->calc(q);
      }
  };

  /** ------------------------------------------------------- **/

  /*
  * A LoopDetector object records all the decisions taken by MVCRA (a channel is assigned to a link)
  * and detects a loop. CausesLoop returns true when a loop is repeated twice:
  * 
  *     ... a b c d e a b c d e a
  *        |---------|---------|
  *                             ↑
  *                           here LoopFound returns true
  */

  class LoopDetector
  {
  public:
    // constructor
    LoopDetector (unsigned int mi) : potential_loop(false), max_iterations(mi) {}
    
    bool CausesLoop (TLinkItrs l, unsigned int ch) throw (int);  // throws an exception if the size of ListOfChanges exceeds max_iterations
    
  private:
    vector<pair<TLinkItrs,unsigned int> > ListOfChanges;
    unsigned int idx_1st_loop, idx_2nd_loop;
    bool potential_loop;
    unsigned int max_iterations;
  };
} /* namespace MVCRA_NS */


using namespace MVCRA_NS;


/** ------------------------------------------------------- **/



/** MVCRA **/


void MVCRA::Configure(const list<pair<string,string> >& params)
{
  // default values
  m_rateadapt = true;
  m_maxnumchanges = 10;

  for (list<pair<string,string> >::const_iterator p = params.begin(); p != params.end(); p++)
  {
    ReadParams RP (p->first, p->second);
    
    RP.read ("rateadapt", m_rateadapt);
    RP.read ("maxnumchanges", m_maxnumchanges);
    RP.read ("ccf", m_ccf);
    RP.read ("lp", m_lp);
    RP.read ("lpp", m_lpp);
    
    if (RP.notFound())
      cerr << endl << "Ignoring unknown attribute (" << p->first << ") for the MVCRA algorithm" << endl << endl;
  }
  m_ld = new LoopDetector (m_maxnumchanges*15);
}


void MVCRA::SetName ()
{
  stringstream ss;
  unsigned int len = string("SumUtotPriority").length();

  ss << "M" << ( m_rateadapt ? "T" : "F" ) << m_maxnumchanges;

  if (m_ccf == "ChannelSetCount")
    ss << "C";
  else if (m_ccf == "DisruptedLinkCount")
    ss << "D";
  else if (m_ccf == "DisruptedLinkFC")
    ss << "F";
  else if (m_ccf == "DisruptExtrLinkCount")
    ss << "E";
  else  // default
    ss << "M";

  if (m_lp == "UtotPriority")
    ss << "U";
  else if (m_lp == "FCPriority")
    ss << "F";
  else // default
  {
    double u0 = 0.5; // default value
    
    if (!m_lp.compare(0,len,"SumUtotPriority") && m_lp.length() > len)
    {
      stringstream ss2(m_lp.substr(len));
      
      if (!(ss2 >> u0))
	u0 = 0.5;
    }
      
    ss << "S" << u0;
  }

  if (m_lpp == "UtotPriority")
    ss << "U";
  else if (!m_lpp.compare(0,len,"SumUtotPriority"))
  {
    double u0 = 0.5; // default value
    
    if (!m_lp.compare(0,len,"SumUtotPriority") && m_lp.length() > len)
    {
      stringstream ss2(m_lp.substr(len));
      
      if (!(ss2 >> u0))
	u0 = 0.5;
    }
      
    ss << "S" << u0;
  }
  else // default
    ss << "F";

  ss >> m_name;
}



/* Given the initial graph, performs a channel assignment
   and returns the induced graph
*/
double MVCRA::Run (Topology* _T)
{
  T = _T;
  m_numchanges = 0;
  ChannelCostFunc* ccf;
  LinkPriority *lp, *lpp;
  unsigned int len = string("SumUtotPriority").length();
  
  if (m_ccf == "ChannelSetCount")
    ccf = new ChannelSetCount;
  else if (m_ccf == "DisruptedLinkCount")
    ccf = new DisruptedLinkCount;
  else if (m_ccf == "DisruptedLinkFC")
    ccf = new DisruptedLinkFC;
  else if (m_ccf == "DisruptExtrLinkCount")
    ccf = new DisruptExtrLinkCount;
  else  // default
    ccf = new MixedCost;

  ccf->SetMVCRA(this);

  if (m_lp == "UtotPriority")
    lp = new UtotPriority(T);
  else if (m_lp == "FCPriority")
    lp = new FCPriority(T);
  else // default
  {
    double u0 = 0.5; // default value
    
    if (!m_lp.compare(0,len,"SumUtotPriority") && m_lp.length() > len)
    {
      stringstream ss(m_lp.substr(len));
      
      if (!(ss >> u0))
	u0 = 0.5;
    }
      
    lp = new SumUtotPriority(T,u0);
  }
  
  if (m_lpp == "UtotPriority")
    lpp = new UtotPriority(T);
  else if (!m_lpp.compare(0,len,"SumUtotPriority"))
  {
    double u0 = 0.5;

    if (m_lpp.length() > len)
    {
      stringstream ss(m_lp.substr(len));
      
      if (!(ss >> u0))
	u0 = 0.5;
    }
    
    lpp = new SumUtotPriority(T,u0);
  }
  else // default
    lpp = new FCPriority(T);

  try
  {
    Q.clear();
    Qp.clear();
    
    changesPerChannel.clear();
    changesPerChannel.resize (T->GetNumNodes(), vector<int>(WMS.numAvailChannels(),0));

    // compute the total utilizations and put all the links into Q
    for (int u = 0; u < T->GetNumNodes(); u++)
      for (TGraphNeighborIterator v = T->Node(u)->NeighborBegin(); v != T->Node(u)->NeighborEnd(); v++)
      {
	SetUtot (v, T->computeUtot(T->LinkItrs(T->Node(u),v)));
	Q.insert(T->LinkItrs(T->Node(u),v));
      }
    
    origChannels.resize(T->GetNumNodes());

    // record the current channels assigned to all the radios (used to detect channel switches)
    for (int u = 0; u < T->GetNumNodes(); u++)
      origChannels[u] = T->Node(u)->Info()->findChannels();


    set<TLinkItrs>::iterator ptr_l;
    bool uFull, vFull;
    int uNumChanges, vNumChanges;
    unsigned int c_selected;

    while (!Q.empty() && m_numchanges < m_maxnumchanges)
    {
      // Extract the link with the highest priority from Q
      ptr_l = max_element (Q.begin(), Q.end(), LinkPriorityPred(lp));
      
      // Erase the extracted link from Q
      TLinkItrs l (*ptr_l);
      Q.erase(ptr_l);

  #ifdef VERBOSE
      cout << "Extracting link " << l.first->Id() << "->" << l.second->Id() << "(" << l.second->RadioChannel() << ")  Utot = " << l.second->UTot() << endl << endl;
  #endif

      // update the Utots before removing the extracted link
      T->updateUtot (l, 0, l.second->Rate());

      // remove the extracted link
      SetRadio (l.second, NULL);

      float new_rate = l.second->Rate();

      set<unsigned int> S, uChannels, vChannels;
      S.clear();

      // put all the channels in S
      for (unsigned int i=1; i<=WMS.numAvailChannels(); S.insert(i++));

      // if RateAdapt is true, the rate is updated
      c_selected = Min_Utot_safe (l, S, new_rate);
      
  #ifdef VERBOSE
      cout << "Selected channel " << c_selected << endl << endl;
  #endif
      
      Change_If (l.first->Info(), c_selected, ccf);
      Change_If (l.second->Info(), c_selected, ccf);

      AssignChannelRate (l, c_selected, new_rate);


  #ifdef VERBOSE
      cout << "The extracted link " << l.first->Id() << "->" << l.second->Id() << " is assigned channel " << l.second->RadioChannel() << " new Utot = " << l.second->UTot() << endl << endl;
  #endif
  #ifdef DEBUG
      if (T->checkUtot(1))
	cout << endl << "checkUtot riuscito al punto A" << endl << endl;
      else
	return 0;
  #endif
      
      /*
      *** Repair the pending links ***
      */
      while (!Qp.empty())
      {
	// Extract the link with the highest priority from Qp
	ptr_l = max_element (Qp.begin(), Qp.end(), LinkPriorityPred(lpp));
	
	// Erase the extracted link from Qp
	TLinkItrs e = *ptr_l;
	Qp.erase (ptr_l);

  #ifdef VERBOSE
      cout << "Extracting pending link " << e.first->Id() << "->" << e.second->Id() << "  Utot = " << e.second->UTot() << endl << endl;
  #endif

	// Find the set of channels of the end nodes of the extracted link
	uChannels = e.first->Info()->findChannels();
	vChannels = e.second->Info()->findChannels();

  #ifdef VERBOSE
	cout << "Channels on " << e.first->Id() << ": ";
	copy(uChannels.begin(), uChannels.end(), ostream_iterator<unsigned int>(cout, " "));
	cout << endl << "Channels on " << e.second->Id() << ": ";
	copy(vChannels.begin(), vChannels.end(), ostream_iterator<unsigned int>(cout, " "));
	cout << endl;
  #endif

	// if a radio is unused, it can be assigned any channel
	if (uChannels.size() < e.first->Info()->numAvailRadios())
	  for (unsigned int i=1; i<=WMS.numAvailChannels(); uChannels.insert(i++));
	if (vChannels.size() < e.second->Info()->numAvailRadios())
	  for (unsigned int i=1; i<=WMS.numAvailChannels(); vChannels.insert(i++));

	// put the common channels of u and v in S
	S.clear();
	set_intersection(uChannels.begin(), uChannels.end(), vChannels.begin(), vChannels.end(), inserter(S, S.begin()));

	if (S.empty())
	{
	  uNumChanges = accumulate (changesPerChannel[e.first->Id()].begin(), changesPerChannel[e.first->Id()].end(), 0);
	  vNumChanges = accumulate (changesPerChannel[e.second->Id()].begin(), changesPerChannel[e.second->Id()].end(), 0);

	  uFull = uNumChanges > (int)e.first->Info()->numAvailRadios();
	  vFull = vNumChanges > (int)e.second->Info()->numAvailRadios();

	  if (!uFull && !vFull)
	    set_union(uChannels.begin(), uChannels.end(), vChannels.begin(), vChannels.end(), inserter(S, S.begin()));
	  else if (uFull && !vFull)
	    S = uChannels;
	  else if (!uFull && vFull)
	    S = vChannels;
	  else
	  {
	    // do not touch the node which experienced the higher number of radio channel changes
	    S = (uNumChanges > vNumChanges ? uChannels : vChannels);

	    cout << "TUTTI E DUE PIENI!!!" << endl << endl;
	  }
	}
	else
	{
	  // NOTE this is OPTIONAL: if the end nodes share a common channel, the link is re-inserted into Q
	  Q.insert(e);
	}
	
  #ifdef VERBOSE
	cout << "Candidate channels: ";
	copy(S.begin(), S.end(), ostream_iterator<unsigned int>(cout, " "));
	cout << endl;
  #endif
      
      // u and v have some channels in common
      c_selected = Min_Utot_safe (e, S, new_rate);    


  #ifdef VERBOSE
	  cout << "Selecting channel " << c_selected << endl << endl;
  #endif

	// these calls immediately return if the node already has the selected channel
	Change_If (e.first->Info(), c_selected, ccf);
	Change_If (e.second->Info(), c_selected, ccf);

	AssignChannelRate (e, c_selected, new_rate);


  #ifdef VERBOSE
      cout << "The pending link " << e.first->Id() << "->" << e.second->Id() << " is repaired using channel " << e.second->RadioChannel() << " new Utot = " << e.second->UTot() << endl << endl;
  #endif
  #ifdef DEBUG
      if (T->checkUtot(1))
	cout << endl << "checkUtot riuscito al punto B" << endl << endl;
      else
	return 0;
  #endif
      
      }

    }
  }
  catch (int)
  {
    cout << "** ERROR ** Maximum number of iterations exceeded. MVCRA quits, likely leaving broken links" << endl;
    m_numchanges = -1;
  }

  return m_numchanges;
}



/** Change_If **/

void MVCRA::Change_If (MeshNode* g, unsigned int c, ChannelCostFunc* ccf)
{
  set<TLinkItrs> *Ldisrupt;
  set<unsigned int> uChannels, vChannels, S;
  
  // if the node already has channel c or an available radio, return
  for (unsigned int i=0; i < g->numAvailRadios(); i++)
    if (g->Radio(i)->Channel() == c || g->Radio(i)->Channel() == 0)
      return;

  unsigned int c_to_erase = Min_Disrupt (g->Id(), c, ccf, &Ldisrupt);

  /* The links in Ldisrupt that still share a common channel after the channel switch can be
   * temporarily repaired without being put in the pending queue nor being extracted from Q.
   * The other links are put into Qp and removed from Q
   */
  for (set<TLinkItrs>::iterator l = Ldisrupt->begin(); l != Ldisrupt->end(); )
  {
    // update the Utots before removing the link
    T->updateUtot (*l, 0, l->second->Rate());

    // remove the link
    SetRadio (l->second, NULL);

    // check whether the end nodes of the link will share a channel after the channel switch
    uChannels = l->first->Info()->findChannels();
    vChannels = l->second->Info()->findChannels();

    // take the channel switch into account
    (l->first->Id() == g->Id() ? uChannels : vChannels).erase(c_to_erase);
    (l->first->Id() == g->Id() ? uChannels : vChannels).insert(c);

    // put the common channels of u and v in S
    S.clear();
    set_intersection(uChannels.begin(), uChannels.end(), vChannels.begin(), vChannels.end(), inserter(S, S.begin()));

    // move the link to Qp if the end nodes do not share channels
    if (S.empty())
    {
      Qp.insert (*l);
      Q.erase (*l);

#ifdef VERBOSE
      cout << "New pending link " << l->first->Id() << "->" << l->second->Id() << "(" << c_to_erase << ")  Utot = " << l->second->UTot() << endl;
#endif
      
      // remove the link from Ldisrupt
      set<TLinkItrs>::iterator temp = l++;
      Ldisrupt->erase(temp);
    }
    else
      l++;
  }
  
  /* The links left in Ldisrupt are repairable. Make the channel switch and repair them */

#ifdef VERBOSE
    cout << "Replacing channel " << c_to_erase << " with channel " << c << " on node " << g->Id() << endl;
#endif

  // replace c_to_erase with c
  for (unsigned int i=0; i<g->numAvailRadios(); i++)
    if (g->Radio(i)->Channel() == c_to_erase)
    {
      SetChannel (g->Radio(i), c);

      if ( origChannels[g->Id()].find(c_to_erase) != origChannels[g->Id()].end() && origChannels[g->Id()].find(c) == origChannels[g->Id()].end() )
	m_numchanges++;
      
      if ( origChannels[g->Id()].find(c_to_erase) == origChannels[g->Id()].end() && origChannels[g->Id()].find(c) != origChannels[g->Id()].end() )
	m_numchanges--;

      break;
    }

  changesPerChannel[g->Id()][c-1]++;

  // temporarily repair the repairable links (that have not been extracted by Q nor put into Qp)
  for (set<TLinkItrs>::iterator l = Ldisrupt->begin(); l != Ldisrupt->end(); l++)
  {
    S = T->haveCommonChannels (l->first->Id(), l->second->Id());

    assert (!S.empty());

    float new_rate;
    unsigned int c_selected = Min_Utot (*l, S, new_rate);

    TLinkItrs nl = *l;
    AssignChannelRate (nl, c_selected, new_rate);


#ifdef VERBOSE
    cout << "Pending link " << nl.first->Id() << "->" << nl.second->Id() << " temporarily repaired using channel " << nl.second->RadioChannel() << " new Utot = " << nl.second->UTot() << endl;
#endif
  }
  
  delete Ldisrupt;

}


/** Min_Utot **/

unsigned int MVCRA::Min_Utot(TLinkItrs l, const std::set< unsigned int >& S, float& new_rate)
{
    map<unsigned int, double> Umax;

    // for each channel c in S
    for (set<unsigned int>::iterator c = S.begin(); c != S.end(); c++)
      Umax[*c] = 0.;

    // for every directed link x->y
    for (int x = 0; x < T->GetNumNodes(); x++)
      for (TGraphNeighborIterator y = T->Node(x)->NeighborBegin(); y != T->Node(x)->NeighborEnd(); y++)
	// whose potential collision domain includes the extracted link and having a channel in S
	if ((T->LinkItrs(T->Node(x),y) != l) && (y->Radio()) && (l.PotInterferesWith(T->LinkItrs(T->Node(x),y))) && (S.find(y->RadioChannel())!=S.end()))
	{
	  if (y->UTot() > Umax[y->RadioChannel()])
	    Umax[y->RadioChannel()] = y->UTot();
	}
    if (l.second->Radio() && S.find(l.second->RadioChannel()) != S.end() && Umax[l.second->RadioChannel()])
      Umax[l.second->RadioChannel()] -= l.second->Flow()/l.second->Rate();

    map<unsigned int, float> rate;
    float save_rate = l.second->Rate();
    double Utot;

    /* we use a fake radio to compute Utot(u->v) for every channel, otherwise the channel of the real radio
    * is changed and all the links using that radio are considered to be on the new channel
    */
    RadioIface* fakeradio = new RadioIface();
    RadioIface* save_radio = l.second->Radio();
    SetRadio(l.second, fakeradio);

    // for each channel
    for (set<unsigned int>::iterator c=S.begin(); c!=S.end(); c++)
    {
      SetChannel(l.second->Radio(), *c);

      if (m_rateadapt)
	SetRate (l.second, Adjust_Rate (l, Umax[*c]));

      rate[*c] = l.second->Rate();

      Utot = T->computeUtot(l) - l.second->Flow()/l.second->Rate();

      if (Utot > Umax[*c])
	Umax[*c] = Utot;

      // in this way, the different rate selected for channels is taken into account in the comparison of Utots
      Umax[*c] += l.second->Flow()/l.second->Rate();
      
      SetRate (l.second, save_rate);
    }

#ifdef VERBOSE
    for (map<unsigned int,double>::iterator mm = Umax.begin(); mm != Umax.end(); mm++)
      cout << "Umax [" << mm->first << "] = " << mm->second << endl;
#endif

    unsigned int c_selected = min_element(Umax.begin(), Umax.end(), compareSecond<unsigned int,double>())->first;
    new_rate = rate[c_selected];
    
    // restore the previous radio
    SetRadio (l.second, save_radio);

    delete fakeradio;

    return c_selected;
}


/** Min_Utot_safe **/

unsigned int MVCRA::Min_Utot_safe(TLinkItrs l, const std::set< unsigned int >& S, float& new_rate)
{
  unsigned int ch = Min_Utot (l, S, new_rate);
  bool loopfound;
  
  try
  {
    loopfound = m_ld->CausesLoop(l,ch);
  }
  catch (int)
  {
    // remove the pending links
    for (set<TLinkItrs>::iterator p = Qp.begin(); p != Qp.end(); p++)
      EraseNeighbor (T->Node(p->first->Id()), p->second);

    // remove the extracted link
    EraseNeighbor (T->Node(l.first->Id()), l.second);

    // rethrow the exception
    throw;
  }

  if (loopfound)
  {
    set<unsigned int> S1(S);

    S1.erase(ch);

    if (S1.empty())
    {
      for (unsigned int i=1; i<=WMS.numAvailChannels(); S1.insert(i++));
      S1.erase(ch);
    }

    ch = Min_Utot (l, S1, new_rate);
  }

  return ch;
}


/** Adjust_Rate **/

float MVCRA::Adjust_Rate(TLinkItrs& l, double Umax)
{
  int m, mmin;
  double min, save;

  float save_rate = l.second->Rate();

  /* determine the max rate for the given link */
  int start_rate = T->MaxAllowLinkRate(l.first->Id(), l.second->Id(), l.second->TxPower(), WMS.min_rate(), WMS.max_rate());

  SetRate (l.second, WMS.availTXrates(start_rate));

  mmin = (m = start_rate);
  min = T->computeUtot(l);

  while ((m>0) && (min > Umax + l.second->Flow()/l.second->Rate()))
  {
    SetRate (l.second, WMS.availTXrates(--m));
    if ( (save = T->computeUtot(l)) < min )
    {
      min = save;
      mmin = m;
    }
  }

  // restore the previous rate
  SetRate (l.second, save_rate);

  return WMS.availTXrates(mmin);
}


/** Min_Disrupt **/

unsigned int MVCRA::Min_Disrupt(int u, unsigned int c, ChannelCostFunc* ccf, set< TLinkItrs >** Ldisrupt)
{
  map<unsigned int,set<TLinkItrs> > W;
  unsigned int c_to_erase;

  // initialize the set of disrupted links for each channel on u
  for (unsigned int i=0; i<T->Node(u)->Info()->numAvailRadios(); i++)
    if (T->Node(u)->Info()->Radio(i)->Channel())
      W[T->Node(u)->Info()->Radio(i)->Channel()].clear();

  for (TGraphNeighborIterator w = T->Node(u)->NeighborBegin(); w != T->Node(u)->NeighborEnd(); w++)
  {
    // only count this link if it is not already pending
    if (w->Radio())
      W[w->RadioChannel()].insert(T->LinkItrs(T->Node(u), w));

    // count the inverse link(s)
    for (TGraphNeighborIterator v = T->Node(w->Id())->NeighborBegin(); v != T->Node(w->Id())->NeighborEnd(); v++)
      if (v->Id() == u && v->Radio())
	W[v->RadioChannel()].insert(T->LinkItrs(T->Node(w->Id()), v));
  }

  ccf->SetNodeChannel(u,c);

  map<unsigned int,double> costmap;
  
  // transform (W.begin(), W.end(), costvec.begin(), bind1st(mem_fun(&ChannelCostFunc::GetCost), ccf));
  
  for (map<unsigned int,set<TLinkItrs> >::iterator mm = W.begin(); mm != W.end(); mm++)
    costmap[mm->first] = ccf->GetCost(*mm);

  c_to_erase = min_element(costmap.begin(), costmap.end(), compareSecond<unsigned int,double>())->first;

  
  if (Ldisrupt)
    *Ldisrupt = new set<TLinkItrs> (W[c_to_erase]);
  
  return c_to_erase;
}





/** LoopDetector method **/

bool LoopDetector::CausesLoop(TLinkItrs l, unsigned int ch) throw (int)
{
  pair<TLinkItrs,unsigned int> lastChange(l,ch);
  vector<pair<TLinkItrs,unsigned int> >::iterator pos;
  
  // if we are not tracking a potential loop, check if one is starting
  if (!potential_loop)
  {
    if ((pos = find (ListOfChanges.begin(), ListOfChanges.end(), lastChange)) != ListOfChanges.end() )
    {
      // this might be a potential loop...
      potential_loop = true;
      idx_1st_loop = pos - ListOfChanges.begin();
      idx_2nd_loop = ListOfChanges.size();
      
    }
  }
  // otherwise, check if the potential loop is confirmed or not
  else
  {
    if ( lastChange != ListOfChanges[idx_1st_loop + (ListOfChanges.size() - idx_2nd_loop)] )
    {
      // false alarm
      potential_loop = false;
    }
    else if ( lastChange == ListOfChanges[idx_2nd_loop] )
    {
      // the detected loop is starting for the third time
      // this change is not recorded and the caller should recompute a different change
      potential_loop = false;
      
      cout << "** ATTENTION **: Assigning channel " << ch << " to link " << l.first->Id() << "->" << l.second->Id() << " causes a loop! " << endl;
      
      return true;
    }
  }
  
  ListOfChanges.push_back(lastChange);
  
  if (ListOfChanges.size() > max_iterations)
    throw 1;
  
  return false;
}


void MVCRA::AssignChannelRate (TLinkItrs& l, unsigned int c, float r)
{
  // check whether another TGraphNeighbor object exists between
  // the same two nodes on the same channel
  TGraphNeighborIterator ol = T->Node(l.first->Id())->FindNeighbor (l.second->Id(), c);

  // if not
  if (ol == T->Node(l.first->Id())->NeighborEnd())
  {
    // update the Utot of the other links
    T->updateUtot (l, c, r);

    // actually change the rate
    SetRate (l.second, r);

    // the selected channel is assigned to the extracted link
    T->AssignChannel2Link(l, c);

    // compute the Utot of link l
    SetUtot (l.second, T->computeUtot(l));
  }
  else
  {
    // update the Utot of the other links
    T->updateUtot (l, c, ol->Rate());

    // add the flow of l on the previous link
    SetFlow (ol, ol->Flow() + l.second->Flow());

    for (FlowDemandsIterator f = l.second->F_e2eBegin(); f != l.second->F_e2eEnd(); f++)
      if (ol->IsDemandPresent (f->first))
	SetFlowDemand (ol, f->first, ol->F_e2e(f->first) + f->second);
      else
	SetFlowDemand (ol, f->first, f->second);

    // remove this neighbor from the topology
    EraseNeighbor (T->Node(l.first->Id()), l.second);

    // if this link is in Q, remove it from Q
    if (Q.find(l) != Q.end())
      Q.erase(l);
    
    l = T->LinkItrs(l.first,ol);
    
    // if the previous link has been already extracted from Q, put it in Q again
    // because its Utot increases after the merge with this link
    if (Q.find(l) == Q.end())
      Q.insert(l);    
    
    // update the Utot of the previous link
    SetUtot (ol, T->computeUtot(l));
  }
}


bool ChannelCostFunc::IsRepairable(const TLinkItrs& l)
{
  set<unsigned int> S;
  
  if (l.first->Id() == u)
//     findChannels (l.second, S)
    S = l.second->Info()->findChannels();
  else
//     findChannels (l.first, S)
    S = l.first->Info()->findChannels();
  
  return (S.find(c_selected) != S.end());
}


void ChannelCostFunc::SetNodeChannel(int _u, unsigned int _c_selected)
{
  u = _u;
  c_selected = _c_selected;
}


const vector<vector<int> >& ChannelCostFunc::getChangesPerChannel ()
{
  return mvcra->changesPerChannel;
}


const std::set< TLinkItrs >& ChannelCostFunc::getQ()
{
  return mvcra->Q;
}


double ChannelSetCount::GetCost (const pair<unsigned int,set<TLinkItrs> >& p)
{
  const vector<int>& v = getChangesPerChannel()[u];
  
  double ratio = accumulate (v.begin(), v.end(), 0);
  
  if (ratio)
    ratio = v[p.first-1] / ratio;
  
  return ratio;
}


double DisruptedLinkCount::GetCost (const pair<unsigned int,set<TLinkItrs> >& p)
{
  int count = 0;
  
  for (set<TLinkItrs>::const_iterator l = p.second.begin(); l != p.second.end(); l++)
    if (!IsRepairable(*l))
      count++;

  return count;
}


double DisruptExtrLinkCount::GetCost (const pair<unsigned int,set<TLinkItrs> >& p)
{
  int count = 0;
  const set<TLinkItrs>& Q = getQ();
  
  for (set<TLinkItrs>::const_iterator l = p.second.begin(); l != p.second.end(); l++)
    if (!IsRepairable(*l) && Q.find(*l) == Q.end())
      count++;

  return count;
}


double DisruptedLinkFC::GetCost(const pair<unsigned int,set<TLinkItrs> >& p)
{
  double total = 0.;
  
  for (set<TLinkItrs>::const_iterator l = p.second.begin(); l != p.second.end(); l++)
    if (!IsRepairable(*l))
      total += l->second->Flow() / l->second->Rate();

  return total;
}


double MixedCost::GetCost (const pair<unsigned int,set<TLinkItrs> >& p)
{
  return (1 + ChannelSetCount::GetCost(p)) * DisruptedLinkFC::GetCost(p);
}


double SumUtotPriority::calc(const TLinkItrs& l)
{
  double ret = 0;

  for (int x = 0; x < T->GetNumNodes(); x++)
    for (TGraphNeighborIterator y = T->Node(x)->NeighborBegin(); y != T->Node(x)->NeighborEnd(); y++) 
      if (y->UTot() >= u0 && y->Radio() && l.second->Radio() && l.IsInTheCollDomainOf(T->LinkItrs(T->Node(x),y))) 
	ret += l.second->Flow() / l.second->Rate() * y->UTot();

  return ret;
}


