#include "wimesh-mpls-ns3.h"
#include "WMutils.h"

#include <set>
#include <map>
#include <cassert>
#include <sstream>

WiMeshMplsNs3::WiMeshMplsNs3()
  : WiMeshNs3 (),
    SplitRatiosVec (),
    disableNhlfe (false),
    out_time (1.0)
{
  stack = "MplsSplitting";
}


void WiMeshMplsNs3::SetDisableNhlfe(bool disable)
{
  disableNhlfe = disable;
}


void WiMeshMplsNs3::SetOutTime(double ot)
{
  out_time = ot;
}


void WiMeshMplsNs3::CreateNs3MeshDevices()
{
  WifiHelper wifi;
  Ipv4AddressHelper address;
  
  NqosWifiMacHelper wifiMac = NqosWifiMacHelper::Default ();

  SetRemoteStationManager (wifi);

//   wifi.SetRemoteStationManager ("ns3::WiMeshWifiManager",
//                                 "WiMeshNs3Ptr", WiMeshNs3PtrValue (WiMeshNs3Ptr (this)));

  wifiMac.SetType ("ns3::AdhocWifiMac");

  /* each element of the map associates a channel with the set of nodes 
     having a radio using that channel */ 
  map<unsigned int, NodeContainer> nodes;
  
  for (TGraphNodeIterator u = topology->NodeBegin (); u != topology->NodeEnd (); u++)
    for (unsigned int r=0; r < u->Info()->numAvailRadios(); r++)
	nodes[u->Info()->Radio(r)->Channel()].Add (meshnodes.Get (u->Id()));

  if (!createRadiosOff)
    // erase channel 0
    nodes.erase (0);

  for (map<unsigned int, NodeContainer>::iterator c = nodes.begin (); c != nodes.end (); c++)
  {
    wifiPhy.Set ("ChannelNumber", UintegerValue (c->first * channelsep));
    meshdevices.Add (wifi.Install (wifiPhy, wifiMac, c->second));
  }

  mpls.EnableInterfaceAutoInstall (meshnodes);

  // assign an IP address to each mesh device
  // the address is 10.1.<channel>.<node_id + 1>
  for (uint32_t i = 0; i < meshnodes.GetN (); ++i)
    for (uint32_t j = 0; j < meshnodes.Get (i)->GetNDevices (); ++j)
    {
      Ptr<WifiNetDevice> wnd = meshnodes.Get (i)->GetDevice (j)->GetObject<WifiNetDevice> ();
      // if it is a WifiNetDevice
      if (wnd != 0)
      {
	stringstream netaddr, hostaddr;
	uint16_t ch = wnd->GetPhy ()->GetChannelNumber () / channelsep;

	// assign a fixed random variable stream number depending on the channel
	wifi.AssignStreams (NetDeviceContainer (wnd), GetRadioRandomVarStreamNumber (i, ch));

	// assign an IP address if the radio is on
	if (ch)
	{
	  netaddr << "10.1." << ch << ".0";
	  hostaddr << "0.0.0." << i+1;

	  address.SetBase (netaddr.str().c_str(), "255.255.255.0", hostaddr.str().c_str());
	  address.Assign (NetDeviceContainer (wnd));
	}
	else
	{
	  // add the device to the set of radios to be put in the sleep mode
	  radiosInitiallyOff.Add (wnd);
	}
      }
    }
 
  mpls.ShowConfig ();
  mpls.DiscoverNetwork ();
  mpls.DisableInterfaceAutoInstall (meshnodes);
}

/* WiMeshMplsNs3::ConfigureForwarding expects to receive the split ratios
 * from the standard input. It is compliant to the format used by
 * SplitRatios::saveRatios
 */
void WiMeshMplsNs3::ConfigureForwarding ()
{
  int src, dst, u, v, ch;
  double f, ratio;
  uint32_t label = 100;
  Ptr<WeightedPolicy> policy;
  
  // abort if the number of demands does not match
  assert (SplitRatiosVec.size() == demands.size ());

  map<int,NodeInfo> info;
  
  // for each demand
  for (unsigned int d = 0; d < demands.size(); d++)
  {
    const DemSplitRatios& dsr = getDemSplitRatios (d);
    
    src = dsr.getEdges().first;
    dst = dsr.getEdges().second;
    f = dsr.getFlow();
    
    assert (CheckDemand (src, dst, f, false));
    
    stringstream saddr, daddr;
    saddr << "192.168." << src << ".1";
    daddr << "192.168." << dst << ".1";

    info.clear ();

    for (int j = 0; j < dsr.getNumRatios(); j++)
    {
      u = dsr.getSplitRatio(j).getLink().first;
      v = dsr.getSplitRatio(j).getLink().second;
      ch = dsr.getSplitRatio(j).getChannel();
      ratio = dsr.getSplitRatio(j).getRatio();

      // if not existing yet, create the NodeInfo for node u
      if (info.find(u) == info.end ())
      {
	info.insert (pair<int,NodeInfo> (u, NodeInfo ()));
	FindChannelIfIndexMap (u, info[u].IfIndex);
      }
      
      // if not existing yet, create the NodeInfo for node v
      if (info.find(v) == info.end ())
      {
	info.insert (pair<int,NodeInfo> (v, NodeInfo ()));
	FindChannelIfIndexMap (v, info[v].IfIndex);
      }

      stringstream hostaddr;
      hostaddr << "10.1." << ch << '.' << v+1;

      // add the Nhlfe to the outgoing_nhlfes vector of u
      info[u].outgoing_nhlfes.push_back (Nhlfe (Swap (label), info[u].IfIndex[ch], 
	Ipv4Address (hostaddr.str ().c_str ())));

      // add the split ratio
      info[u].split_ratios.push_back (ratio);
      
      // add the current label to the incoming_labels set of v
      info[v].incoming_labels.insert (label++);
    }
    
    for (map<int,NodeInfo>::iterator m = info.begin (); m != info.end (); ++m)
    {
      //MplsSwitch sw (meshnodes.Get (m->first));
      Ptr<MplsNode> sw = DynamicCast<MplsNode> (meshnodes.Get (m->first));
      
      // if the current node has outgoing nhlfes (all nodes but the destination)
      if (!m->second.outgoing_nhlfes.empty ())
      {
	policy = Create<WeightedPolicy> ();
	policy->SetWeights (m->second.split_ratios);
	
	// if the current node is the source of the demand
	if (m->first == src)
	{
	  // add an FTN entry
	  Ptr<FecToNhlfe> ftn = Create<FecToNhlfe> (Fec::Build (Ipv4Source (saddr.str ().c_str ()) &&
	    Ipv4Destination (daddr.str ().c_str ())), m->second.outgoing_nhlfes[0], policy);
	  
	  for (unsigned int i = 1; i < m->second.outgoing_nhlfes.size (); ++i)
	    ftn->AddNhlfe (m->second.outgoing_nhlfes[i]);
	  
	  sw->GetFtnTable ()->push_back (ftn);
	}
	else
	{
	  // add an ILM entry for each incoming label
	  for (set<uint32_t>::iterator l = m->second.incoming_labels.begin (); l != m->second.incoming_labels.end (); ++l)
	  {
	    Ptr<IncomingLabelMap> ilm = Create<IncomingLabelMap> (*l, m->second.outgoing_nhlfes[0], policy);
	    
	    for (unsigned int i = 1; i < m->second.outgoing_nhlfes.size (); ++i)
	      ilm->AddNhlfe (m->second.outgoing_nhlfes[i]);
	    
	    sw->GetIlmTable ()->push_back (ilm);
	  }
	}
      }
      // else, the current node is the destination
      else
      {
	assert (m->first == dst);
	
	NhlfeSelectionPolicyHelper policy;
	
	for (set<uint32_t>::iterator l = m->second.incoming_labels.begin (); l != m->second.incoming_labels.end (); ++l)
	{
	  Ptr<IncomingLabelMap> ilm = Create<IncomingLabelMap> (*l, Nhlfe (Pop ()), policy.Create ());
	  
	  sw->GetIlmTable ()->push_back (ilm);
	}
      }
    }
  }
}

void
WiMeshMplsNs3::PrintConfiguration ()
{
  for (uint32_t n = 0; n < meshnodes.GetN(); ++n)
  {
    cout << endl << "Node " << n << endl;
    
    MplsNode::IlmTable* ilm = DynamicCast<MplsNode> (meshnodes.Get (n))->GetIlmTable ();
    
    if (ilm)
      for (MplsNode::IlmTable::iterator i = ilm->begin (); i != ilm->end (); ++i)
	(*i)->Print (cout);
      
    MplsNode::FtnTable* ftn = DynamicCast<MplsNode> (meshnodes.Get (n))->GetFtnTable ();
    
    if (ftn)
      for (MplsNode::FtnTable::iterator i = ftn->begin (); i != ftn->end (); ++i)
	(*i)->Print (cout);
  }
}


void WiMeshMplsNs3::FindChannelIfIndexMap (int u, map<uint16_t, uint32_t>& m)
{
  Ptr<Mpls> mpls = meshnodes.Get (u)->GetObject<Mpls> ();

  assert (mpls != 0);

  // find the mapping between channel and Id of the interface using that channel
  for (uint32_t n = 0; n < meshnodes.Get (u)->GetNDevices(); n++)
  {
    Ptr<WifiNetDevice> wnd = meshnodes.Get (u)->GetDevice (n)->GetObject<WifiNetDevice> ();
    // if it is a WifiNetDevice and it is not in sleep mode (channel 0)
    if (wnd != 0 && wnd->GetPhy ()->GetChannelNumber () )
      m[wnd->GetPhy ()->GetChannelNumber () / channelsep] = mpls->GetInterfaceForDevice(wnd)->GetIfIndex ();
  }
}

void
WiMeshMplsNs3::DoCreateNodes ()
{
  meshnodes = mpls.CreateAndInstall (topology->GetNumNodes());
}

void
WiMeshMplsNs3::DoReceiveFinalDataFailed(int me, Mac48Address neighbor)
{
  if (!disableNhlfe)
    return;
  
  Ptr<WeightedPolicy> policy;
  int channel;
  Ptr<WifiNetDevice> neighdevice;
  int neigh = DoFindNeighborByMacAddress (me, neighbor, channel, &neighdevice);

  // find the Ipv4 address of the device of the neighbor
  Ptr<Ipv4> ipv4 = meshnodes.Get (neigh)->GetObject<Ipv4> ();
  
  Ipv4Address neighaddr = ipv4->GetAddress (ipv4->GetInterfaceForDevice (neighdevice), 0).GetLocal ();

  // get the FTN table
  MplsNode::FtnTable* ftn = DynamicCast<MplsNode> (meshnodes.Get (me))->GetFtnTable ();
  
  // browse all the FTN rules
  for (list<Ptr<FecToNhlfe> >::iterator i = ftn->begin (); i != ftn->end (); ++i)
    // browse all the associated NHLFEs
    for (uint32_t n = 0; n < (*i)->GetNNhlfe (); n++)
      if (Ipv4Address::IsMatchingType ((*i)->GetNhlfe (n).GetNextHop ()) && Ipv4Address::ConvertFrom ((*i)->GetNhlfe (n).GetNextHop ()) == neighaddr)
      {
	policy = DynamicCast<WeightedPolicy> ((*i)->GetPolicy ());
	if (policy)
	{
	  policy->Deactivate (n);

	  // schedule its reactivation
	  Simulator::Schedule (Seconds (out_time), &WeightedPolicy::Reactivate, policy, n);
	}
      }

  // get the ILM table
  MplsNode::IlmTable* ilm = DynamicCast<MplsNode> (meshnodes.Get (me))->GetIlmTable ();
  
  // browse all the ILM rules
  for (list<Ptr<IncomingLabelMap> >::iterator i = ilm->begin (); i != ilm->end (); ++i)
    // browse all the associated NHLFEs
    for (uint32_t n = 0; n < (*i)->GetNNhlfe (); n++)
      if (Ipv4Address::IsMatchingType ((*i)->GetNhlfe (n).GetNextHop ()) && Ipv4Address::ConvertFrom ((*i)->GetNhlfe (n).GetNextHop ()) == neighaddr)
      {
	policy = DynamicCast<WeightedPolicy> ((*i)->GetPolicy ());
	if (policy)
	{
	  policy->Deactivate (n);

	  // schedule its reactivation
	  Simulator::Schedule (Seconds (out_time), &WeightedPolicy::Reactivate, policy, n);
	}
      }
}

void
WiMeshMplsNs3::ReadSplitRatiosFromFile (istream& is)
{
  int NDemands, src, dst, Nsplit, u, v, ch;
  double f, ratio;
  
//   skipComment (is);
//   is >> NDemands;
  readValue (is, NDemands, "the number of demands");
  
  // for each demand
  for (int d = 0; d < NDemands; d++)
  {
    is >> src >> dst >> f >> Nsplit;
    
    DemSplitRatios dsr(pair<int,int>(src,dst), f);
    
    for (int j = 0; j < Nsplit; j++)
    {
      is >> u >> v >> ch >> ratio;
      
      dsr.AddSplitRatio(TSplitRatio(pair<int,int>(u,v),ch,ratio));
    }
    
    SplitRatiosVec.push_back(dsr);
  }
}

void
WiMeshMplsNs3::CalcSplitRatiosFromTopology ()
{
  vector<DemSplitRatios>::iterator d;
  
  if (!topology)
  {
    cerr << "ERROR: CalcSplitRatiosFromTopology: the topology is not set" << endl;
    return;
  }

  /* check that e2e flows are consistent and obtain the set of demands */
  map<pair<int,int>,double> demands, outflow;
  
  if (!(topology->checkE2Eflows(&demands)))
    cerr << endl << "WARNING: CalcSplitRatiosFromTopology: Checking E2E flows FAILED" << endl << endl;

  // insert the demands into the vector of split ratios
  for (map<pair<int,int>,double>::iterator m = demands.begin(); m != demands.end(); m++)
    SplitRatiosVec.push_back(DemSplitRatios(pair<int,int>(m->first.first,m->first.second),m->second));
  
  for (TGraphNodeIterator u = topology->NodeBegin(); u != topology->NodeEnd(); u++)
  {
    for (map<pair<int,int>,double>::iterator dem = demands.begin(); dem != demands.end(); dem++)
      outflow[dem->first] = 0.;

    // compute the flow outgoing node u for each demand
    for (TGraphNeighborIterator v = u->NeighborBegin(); v != u->NeighborEnd(); v++)
      for (FlowDemandsIterator f = v->F_e2eBegin(); f != v->F_e2eEnd(); f++)
	outflow[f->first] += f->second;

    // compute the split ratios of the links outgoing node u
    for (TGraphNeighborIterator v = u->NeighborBegin(); v != u->NeighborEnd(); v++)
      for (FlowDemandsIterator f = v->F_e2eBegin(); f != v->F_e2eEnd(); f++)
      {
	// find the corresponding demand in SplitRatiosVec
	for (d = SplitRatiosVec.begin(); d != SplitRatiosVec.end() && d->getEdges() != f->first; d++);

	assert (d != SplitRatiosVec.end());

	if ((outflow[f->first] && f->second > 0.0001) || disableNhlfe)
	  d->AddSplitRatio(TSplitRatio(pair<int,int>(u->Id(),v->Id()),v->RadioChannel(), outflow[f->first] ? f->second/outflow[f->first] : 0.));
      }
  }
}


void WiMeshMplsNs3::ChangeTopology(Topology* T, int* changes)
{
  /*
   * switch channels and update IP addresses
   */
  vector<multiset<unsigned int> > newChSet, oldChSet;

  int res = topology->countRadioChanged (T, oldChSet, newChSet);
  
  if (changes)
    *changes = res;
  
  for (int u = 0; u < (int)T->GetNumNodes(); u++)
  {
    // do the channel switches and update the IP addresses
    multiset<unsigned int>::iterator newCh = newChSet[u].begin();
    for (multiset<unsigned int>::iterator oldCh = oldChSet[u].begin(); oldCh != oldChSet[u].end(); oldCh++, newCh++)
    {
      DoChannelSwitch (u, *oldCh, *newCh);
      
      // update the IP address: 10.1.<new_channel>.<node_id + 1>
      Ptr<Ipv4> ipv4 = meshnodes.Get (u)->GetObject<Ipv4> ();
      
      assert (ipv4);
      
      stringstream oldAddr, newAddr;
      oldAddr << "10.1." << *oldCh << "." << u+1;
      newAddr << "10.1." << *newCh << "." << u+1;
      
      int32_t interface = ipv4->GetInterfaceForAddress (oldAddr.str().c_str());
      ipv4->RemoveAddress (interface, 0);
      ipv4->AddAddress (interface, Ipv4InterfaceAddress (newAddr.str().c_str(), "255.255.255.0") );
    }
  }

  // the topology pointed to by WiMeshNs3 is the new one (required to call CalcSplitRatiosFromTopology)
  // Keeping updated the topology member is also required because, after a channel switch,
  // the WiMeshWifiManager tries to find the link rate. If the WiMesh topology is not updated,
  // a TGraphNeighbor with the new channel is not found and the basic rate (6Mbps) is used
  topology = T;
  
  // compute the new set of split ratios
  SplitRatiosVec.clear();
  CalcSplitRatiosFromTopology();

  // clear the MPLS tables
  for (int u = 0; u < (int)topology->GetNumNodes(); u++)
  {
    Ptr<MplsNode> sw = DynamicCast<MplsNode> (meshnodes.Get (u));
    
    MplsNode::FtnTable* ftn = sw->GetFtnTable();
    if (ftn)
      ftn->clear();
    
    MplsNode::IlmTable* ilm = sw->GetIlmTable();
    if (ilm)
      ilm->clear();
  }

  mpls.DiscoverNetwork ();
  
  // re-configure forwarding
  ConfigureForwarding();
}
