/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2010 Network Security Lab, University of Washington, Seattle.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors: Sidharth Nabar <snabar@uw.edu>, He Wu <mdzz@u.washington.edu>
 *          Stefano Avallone <stavallo@unina.it>
 */

#ifndef WIMESH_NODE_ENERGY_MODEL_H
#define WIMESH_NODE_ENERGY_MODEL_H

#include "ns3/device-energy-model.h"
#include "ns3/nstime.h"
#include "ns3/event-id.h"
#include "ns3/traced-value.h"
#include "ns3/wifi-phy.h"
#include "ns3/net-device-container.h"
#include <vector>

namespace ns3 {

/**
 * \ingroup energy
 * A WifiPhy listener class for notifying the WiMeshNodeEnergyModel of Wifi radio
 * state change.
 *
 */
class WiMeshNodeEnergyModelPhyListener : public WifiPhyListener
{
public:
  WiMeshNodeEnergyModelPhyListener ();
  virtual ~WiMeshNodeEnergyModelPhyListener ();

  /**
   * \brief Sets the change state callback. Used by helper class.
   *
   * \param callback Change state callback.
   */
  void SetChangeStateCallback (DeviceEnergyModel::ChangeStateCallback callback);

  /**
   * \param duration the expected duration of the packet reception.
   *
   * Defined in ns3::WifiPhyListener, does nothing here
   */
  virtual void NotifyRxStart (Time duration)
  {
  }

  /**
   * Defined in ns3::WifiPhyListener, does nothing here
   */
  virtual void NotifyRxEndOk (void)
  {
  }

  /**
   * Defined in ns3::WifiPhyListener, does nothing here
   */
  virtual void NotifyRxEndError (void)
  {
  }

  /**
   * \param duration the expected transmission duration.
   * \param txPowerDbm the nominal tx power in dBm
   *
   * Defined in ns3::WifiPhyListener, does nothing here
   */
  virtual void NotifyTxStart (Time duration, double txPowerDbm)
  {
  }

  /**
   * \param duration the expected busy duration.
   *
   * Defined in ns3::WifiPhyListener, does nothing here
   */
  virtual void NotifyMaybeCcaBusyStart (Time duration)
  {
  }

  /**
   * \param duration the expected channel switching duration.
   *
   * Defined in ns3::WifiPhyListener, does nothing here
   */
  virtual void NotifySwitchingStart (Time duration)
  {
  }

  /**
   * Defined in ns3::WifiPhyListener
   */
  virtual void NotifySleep (void);

  /**
   * Defined in ns3::WifiPhyListener
   */
  virtual void NotifyWakeup (void);

private:
  /**
   * Change state callback used to notify the WiMeshNodeEnergyModel of a state
   * change.
   */
  DeviceEnergyModel::ChangeStateCallback m_changeStateCallback;
};

// -------------------------------------------------------------------------- //

/**
 * \ingroup energy
 * \brief A WiMesh node energy model.
 * 
 * 2 states are defined for a WiMesh node: ACTIVE (if at least a radio is not in
 * sleep mode) and STANDBY (if all the radios are in sleep mode).
 * To know the state of the radios, a listener for each radio must be added. This
 * is done by the AddNetDevice method.
 * 
 * Energy calculation: Every time a radio goes to sleep or is woken up, the 
 * associated PHY listener notifies this model (through the ChangeState callback),
 * which notifies in turn the EnergySource object. The EnergySource object will
 * query this model for the total current. Then the EnergySource object uses the
 * total current to calculate energy.
 */
class WiMeshNodeEnergyModel : public DeviceEnergyModel
{
public:
  /**
   * The state of the WiMesh node.
   */
  enum State
  {
    /**
     * The node is active.
     */
    ACTIVE,
    /**
     * The node is paused.
     */
    STANDBY
  };

  static TypeId GetTypeId (void);
  WiMeshNodeEnergyModel ();
  virtual ~WiMeshNodeEnergyModel ();

  /**
   * \brief Sets pointer to EnergySouce installed on node.
   *
   * \param source Pointer to EnergySource installed on node.
   *
   * Implements DeviceEnergyModel::SetEnergySource.
   */
  virtual void SetEnergySource (Ptr<EnergySource> source);

  /**
   * \brief Adds a device (radio interface) to this WiMesh node.
   *
   * \param device Pointer to the device installed on node.
   */
  virtual void AddNetDevice (Ptr<NetDevice> device);

  /**
   * \returns Total energy consumption of the wifi device.
   *
   * Implements DeviceEnergyModel::GetTotalEnergyConsumption.
   */
  virtual double GetTotalEnergyConsumption (void) const;

  /**
   * \returns The time spent in standby mode by the WiMesh node.
   */
  virtual double GetStandbyTime (void) const;

  // Setter & getters for state power consumption.
  double GetActiveCurrentA (void) const;
  void SetActiveCurrentA (double activeCurrentA);
  double GetStandbyCurrentA (void) const;
  void SetStandbyCurrentA (double standbyCurrentA);

  /**
   * \brief Changes state of the WifiRadioEnergyMode.
   *
   * \param newState New state the wifi radio is in.
   *
   * Implements DeviceEnergyModel::ChangeState.
   */
  virtual void ChangeState (int newState);

  /**
   * \brief Handles energy depletion.
   *
   * Not implemented
   */
  virtual void HandleEnergyDepletion (void)
  {
  }

  /**
   * \brief Handles energy recharged.
   *
   * Not implemented
   */
  virtual void HandleEnergyRecharged (void)
  {
  }


private:
  void DoDispose (void);

  /**
   * \returns Current draw of device, at current state.
   *
   * Implements DeviceEnergyModel::GetCurrentA.
   */
  virtual double DoGetCurrentA (void) const;

  /**
   * \param state New state the radio device is currently in.
   *
   * Sets current state. This function is private so that only the energy model
   * can change its own state.
   */
  void SetWiMeshNodeState (const WiMeshNodeEnergyModel::State state);

private:
  Ptr<EnergySource> m_source;

  // Member variables for current draw in different modes.
  double m_activeCurrentA;
  double m_standbyCurrentA;

  // This variable keeps track of the total energy consumed by this model.
  TracedValue<double> m_totalEnergyConsumption;

  Time m_standbyTime;      // the time spent in standby mode

  // State variables.
  WiMeshNodeEnergyModel::State m_currentState;  // current state the node is in
  Time m_lastUpdateTime;          // time stamp of previous energy update

  uint8_t m_nPendingChangeState;
  bool m_isSupersededChangeState;

  // A WifiPhy listener for each radio
  std::vector<WiMeshNodeEnergyModelPhyListener*> m_listeners;

  // set of radios
  NetDeviceContainer m_devices;
};

} // namespace ns3

#endif /* WIMESH_NODE_ENERGY_MODEL_H */
