#include "wimesh-ns3.h"
#include "WMutils.h"
#include "udp-trace-client.h"
#include "ns3/energy-module.h"
#include "wimesh-node-energy-model.h"
#include "ns3/point-to-point-module.h"
#include "ns3/propagation-module.h"
#include <sstream>
#include <cassert>
#include <cstdarg>

WiMeshNs3::WiMeshNs3 ()
  : topology (0),
    wifiPhy (YansWifiPhyHelper::Default ()),
    wifiChannel (YansWifiChannelHelper::Default ()),
    ns3seed (1),
    ns3run (1),
    channelsep (4),
    createRadiosOff (false),
    trafficType (string("OnOff")),
    transportProto (string("ns3::UdpSocketFactory")),
    t_on_min (0.5),
    t_on_max (1.5),
    t_off_min (0.25),
    t_off_max (0.75),
    traceFiles (),
    packetsize (1400),
    duration (60.0),
    RtsCtsThreshold (-1),
    pcapEnabled (false),
    failures (),
    ch_switches(),
    enableFlowMonitor (false),
    measure_interval (1.0),
    measure_perflow (false),
    enableEnergyMeter (false),
    initEnergySeed (RndVar::getSeed()),
    energyHarvesters (NULL),
    RSMname (string("ns3::WiMeshWifiManager"))
{
  stringstream ss(string("C 100000.0"));
  initEnergy = new RndVar (ss);
}

WiMeshNs3::WiMeshNs3 (const WiMeshNs3& w)
{
  *this = w;
//   DoInit ();
}

void WiMeshNs3::DoInit()
{
  RngSeedManager::SetSeed (ns3seed);
  RngSeedManager::SetRun (ns3run);

  Config::SetDefault ("ns3::YansWifiPhy::TxGain", DoubleValue (RadioIface::default_db_gain));
  Config::SetDefault ("ns3::YansWifiPhy::RxGain", DoubleValue (RadioIface::default_db_gain));

  //Configure the WiMesh available power levels
  Config::SetDefault ("ns3::YansWifiPhy::TxPowerStart", DoubleValue (WMS.minTxPowerLevel()));
  Config::SetDefault ("ns3::YansWifiPhy::TxPowerEnd", DoubleValue (WMS.maxTxPowerLevel()));
  Config::SetDefault ("ns3::YansWifiPhy::TxPowerLevels", UintegerValue(WMS.numTxPowerLevels()));

  Config::SetDefault ("ns3::TcpSocket::SegmentSize", UintegerValue (packetsize));

  if (RtsCtsThreshold >= 0)
    Config::SetDefault ("ns3::WifiRemoteStationManager::RtsCtsThreshold", UintegerValue (RtsCtsThreshold));

  // increase the maximum number of peers in dot11sStack
  Config::SetDefault("ns3::dot11s::PeerManagementProtocol::MaxNumberOfPeerLinks", UintegerValue (70));
  
  // Initializes Channel and Phy
  Ptr<YansWifiChannel> c = wifiChannel.Create ();
  c->AssignStreams (GetChannelRandomVarStreamNumber ());

  // Set the propagation loss model stored in the WiMesh topology
  ObjectFactory factory;
  string modelname = "ns3::" + string(WMS.PropagationModel()->Name()) + "PropagationLossModel";
  factory.SetTypeId (modelname);
  // set attributes
  for (vector<pair<string,double> >::const_iterator p = WMS.PropagationModel()->Begin(); p != WMS.PropagationModel()->End(); p++)
  {
    factory.Set (p->first, DoubleValue (p->second));
  }
  Ptr<ns3::PropagationLossModel> ploss = factory.Create ()->GetObject<ns3::PropagationLossModel>();

  c->SetPropagationLossModel (ploss);

  wifiPhy.SetChannel (c);
  wifiPhy.SetPcapDataLinkType (YansWifiPhyHelper::DLT_IEEE802_11_RADIO); 
}


void WiMeshNs3::SetTopology(Topology* _topology)
{
  topology = _topology;
}

Topology* WiMeshNs3::GetTopology()
{
  return topology;
}

void WiMeshNs3::SetChannelSep (unsigned int cs)
{
  channelsep = cs;
}

unsigned int WiMeshNs3::GetChannelSep()
{
  return channelsep;
}

void WiMeshNs3::SetTrafficType (std::string tt)
{
  trafficType = tt;
}

void WiMeshNs3::AddTraceFile (std::string f)
{
  traceFiles.push_back (f);
}

void WiMeshNs3::SetTransportProto (std::string tp)
{
  transportProto = tp;
}


void WiMeshNs3::SetTOn(double min, double max)
{
  t_on_min = min;
  t_on_max = max;
}


void WiMeshNs3::SetTOff(double min, double max)
{
  t_off_min = min;
  t_off_max = max;
}

void WiMeshNs3::GetTOn(double& min, double& max)
{
  min = t_on_min;
  max = t_on_max;
}

void WiMeshNs3::GetTOff(double& min, double& max)
{
  min = t_off_min;
  max = t_off_max;
}

void WiMeshNs3::SetPacketSize(unsigned int ps)
{
  packetsize = ps;
}

unsigned int WiMeshNs3::GetPacketSize()
{
  return packetsize;
}

void WiMeshNs3::SetDuration(double d)
{
  duration = d;
}

double WiMeshNs3::GetDuration()
{
  return duration;
}

void WiMeshNs3::SetPcapEnabled(bool enable)
{
  pcapEnabled = enable;
}

void WiMeshNs3::SetFlowMonitorEnabled(bool flowmon)
{
  enableFlowMonitor = flowmon;
}

void WiMeshNs3::SetEnergyMeterEnabled(bool emeter)
{
  enableEnergyMeter = emeter;
}

void WiMeshNs3::SetInitialEnergy(RndVar* source)
{
  initEnergy = source;
}

void WiMeshNs3::SetInitialEnergySeed(double seed)
{
  initEnergySeed = seed;
}

WiMeshNs3::FailureEvent::FailureEvent (double _time, double _noise, int _type, ...)
  : time (_time),
    noise (_noise),
    type (_type)
{
  va_list params;
  va_start (params, _type);
  
  switch (type)
  {
    case 1:
      node = va_arg (params, int);
      channel = va_arg (params, int);
      break;
      
    case 2:
      radio = va_arg (params, int);
      break;
      
    case 3:
      node = va_arg (params, int);
      break;
      
    default:
      throw;
  }
}

int WiMeshNs3::GetWiMeshRadioTXpower (unsigned int node, unsigned int channel)
{
  TGraphNeighborIterator v = topology->Node(node)->NeighborBegin();
  for ( ; v != topology->Node(node)->NeighborEnd() && v->RadioChannel() != channel; v++);

  if (v != topology->Node(node)->NeighborEnd())
    return v->TxPower();

  // print a warning if the radio is not turned off (channel 0)
  if (channel)
    cerr << endl << "** WARNING: No link outgoing node " << node << "  using channel " << channel 
         << " has been found in the WiMesh topology **" << endl << endl;

  return WMS.defaultTxPowerLevel();
}

WiMeshNs3::Ns3Attribute::Ns3Attribute (string n, Ptr<AttributeValue> p)
  : m_name (n),
    m_ptr_value (p)
{
}

void WiMeshNs3::SetRSMoptions ()
{
  Ptr<WifiNetDevice> wnd;
  Ptr<WifiRemoteStationManager> rsm;

  for (NodeContainer::Iterator n = meshnodes.Begin(); n != meshnodes.End(); n++)
    for (uint32_t i = 0; i < (*n)->GetNDevices (); ++i)
    {
      wnd = (*n)->GetDevice (i)->GetObject<WifiNetDevice> ();
      // if it is a WifiNetDevice
      if (wnd != 0)
      {
	unsigned int channel = wnd->GetPhy ()->GetChannelNumber () / channelsep;

	rsm = wnd->GetRemoteStationManager ();

	if (!rsm)         // no wifi remote station manager ??
	{
	  cerr << endl << "** WARNING: No Wifi remote station manager found on the radio using channel " << channel
	       << " on node " << (*n)->GetId() << " **" << endl << endl;
	  continue;
	}

	// set the options specified in the XML config file
	for (vector<Ns3Attribute>::iterator i = RSMoptions.begin(); i != RSMoptions.end(); i++)
	  rsm->SetAttribute (i->m_name, *PeekPointer(i->m_ptr_value));

	// set the TX power level for this WifiNetDevice
	rsm->SetDefaultTxPowerLevel (GetWiMeshRadioTXpower((*n)->GetId(), channel));
      }
    }
}

WiMeshNs3::Ns3EnergyHarvesters::Ns3EnergyHarvesters (string vartype, vector<RndVar*> params, vector<double> seeds, double interval)
  : m_vartype (vartype), m_params (params), m_interval (interval)
{
  for (int i = 0; i < m_params.size(); i++)
    urng.push_back (MotherOfAll(seeds[i]));
}

string WiMeshNs3::Ns3EnergyHarvesters::GetDistribution ()
{
  stringstream ss;

  ss << "ns3::" << m_vartype << "RandomVariable";

  if (m_vartype == "Constant")
  {
    Random::Set(urng[0]);
    ss << "[Constant=" << m_params[0]->rv->Next() << "]";
  }
  else if (m_vartype == "Uniform")
  {
    // first parameter is the average, second one is the width
    Random::Set(urng[0]);
    double avg = m_params[0]->rv->Next();

    Random::Set(urng[1]);
    double width = m_params[1]->rv->Next();

    ss << "[Min=" << avg-width/2 << "|Max=" << avg+width/2 << "]";
  }
  else if (m_vartype == "Exponential")
  {
    Random::Set(urng[0]);
    ss << "[Mean=" << m_params[0]->rv->Next() << "]";
  }

  return ss.str();
}

void NullFunction ()
{}

void WiMeshNs3::DoAttachEnergySources ()
{
  BasicEnergySourceHelper basicSourceHelper;
  Ptr<WifiNetDevice> wnd;
  MotherOfAll urng(initEnergySeed);    // declare uniform random number generator
  Random::Set(urng);                   // set urng as generator to be used

  // install an energy source on each node
  ESnodes = basicSourceHelper.Install (meshnodes);

  for (uint32_t n = 0; n < meshnodes.GetN (); n++)
  {
    double voltage = ESnodes.Get(n)->GetSupplyVoltage();
    MeshNode *mi = topology->Node(n)->Info();

    // set the initial energy of the energy source
    DynamicCast<BasicEnergySource>(ESnodes.Get(n))->SetInitialEnergy (initEnergy->rv->Next());

    // attach an energy harvester to the source, if specified
    if (energyHarvesters)
    {
      BasicEnergyHarvesterHelper basicHarvesterHelper;

      if (energyHarvesters->GetInterval() > 0.)
	basicHarvesterHelper.Set ("PeriodicHarvestedPowerUpdateInterval", TimeValue (Seconds (energyHarvesters->GetInterval())));

      basicHarvesterHelper.Set ("HarvestablePower", StringValue(energyHarvesters->GetDistribution ()));

      // install harvester on the energy source
      EnergyHarvesterContainer ehc = basicHarvesterHelper.Install (ESnodes.Get(n));

      DynamicCast<BasicEnergyHarvester>(ehc.Get(0))->AssignStreams (GetHarvesterRandomVarStreamNumber(n));
    }

    // create the WiMesh node energy model used to simulate the node power consumption
    ObjectFactory factory;
    factory.SetTypeId ("ns3::WiMeshNodeEnergyModel");
    // set attributes (base and standby power consumption)
    factory.Set ("ActiveCurrentA", DoubleValue (mi->GetBasePowerConsumption() / voltage));
    factory.Set ("StandbyCurrentA", DoubleValue (mi->GetStandbyPowerConsumption() / voltage));
    Ptr<WiMeshNodeEnergyModel> wnem = factory.Create ()->GetObject<WiMeshNodeEnergyModel>();
    // set energy source pointer
    wnem->SetEnergySource (ESnodes.Get(n));
    // add wnem to device model list in energy source
    ESnodes.Get(n)->AppendDeviceEnergyModel (wnem);

    for (uint32_t i = 0; i < meshnodes.Get(n)->GetNDevices (); ++i)
    {
      wnd = meshnodes.Get(n)->GetDevice (i)->GetObject<WifiNetDevice> ();
      // if it is a WifiNetDevice
      if (wnd != 0)
      {
	int channel = wnd->GetPhy ()->GetChannelNumber () / channelsep;

	// find the radio with such a channel
	int ri = mi->findRadioWithChannel(channel);

	assert ( ri >= 0 );

	WifiRadioEnergyModelHelper radioEnergyHelper;

	radioEnergyHelper.Set("SleepCurrentA", DoubleValue (mi->Radio(ri)->Rho_Sleep() / voltage) );
	radioEnergyHelper.Set("IdleCurrentA", DoubleValue (mi->Radio(ri)->Rho_Idle() / voltage) );
	radioEnergyHelper.Set("CcaBusyCurrentA", DoubleValue (mi->Radio(ri)->Rho_Idle() / voltage) );
	radioEnergyHelper.Set("SwitchingCurrentA", DoubleValue (mi->Radio(ri)->Rho_Idle() / voltage) );
	radioEnergyHelper.Set("RxCurrentA", DoubleValue (mi->Radio(ri)->Rho_Rx() / voltage) );

	foRadioTxPowerLinear* rho_tx_linear = dynamic_cast<foRadioTxPowerLinear*> (mi->Radio(ri)->Rho_Tx());

	if (rho_tx_linear)
	{
	  if (rho_tx_linear->getA() == 0)
	  // the absorbed power in transmission mode is constant
	  {
	    radioEnergyHelper.Set("TxCurrentA", DoubleValue (rho_tx_linear->getB() / voltage) );
	  }
	  else
	  // the absorbed power in tx mode is a linear function of the nominal tx power
	  {
	    radioEnergyHelper.SetTxCurrentModel ("ns3::LinearWifiTxCurrentModel",
                                                 "Voltage", DoubleValue (voltage),
                                                 "IdleCurrent", DoubleValue (rho_tx_linear->getB() / voltage),
                                                 "Eta", DoubleValue (1. / rho_tx_linear->getA()));
	  }
	}
	else
	  cerr << endl << "** WARNING: Unknown radio power consumption function object **" << endl << endl;

// 	radioEnergyHelper.Set("TxCurrentA", DoubleValue ((*mi->Radio(ri)->Rho_Tx())(DbmToW(WMS.TxPowerLevel(GetWiMeshRadioTXpower((*n)->GetId(), channel)))) / voltage) );
// 	radioEnergyHelper.Set("WiMeshNs3Ptr", WiMeshNs3PtrValue (WiMeshNs3Ptr (this)));

	if (!channel)
	  // if channel is 0, the radio is put in sleep mode and should not be woken up
	  radioEnergyHelper.SetRechargedCallback (MakeCallback(&NullFunction));

	// this device draws power from the energy source attached to the node
	radioEnergyHelper.Install (wnd, ESnodes.Get(n));

	// add the device to the WiMesh node energy model
	wnem->AddNetDevice (wnd);
      }
    }
  }
}


void WiMeshNs3::GetNodesInitialEnergy (vector<double>* energy)
{
  Simulator::Schedule (Seconds (0.), &WiMeshNs3::DoGetNodesInitialEnergy, this, energy);
}


void WiMeshNs3::DoGetNodesInitialEnergy (vector<double>* energy)
{
  energy->resize(ESnodes.GetN());

  for (uint32_t i = 0; i < ESnodes.GetN(); i++)
    energy->at(i) = ESnodes.Get(i)->GetInitialEnergy();
}


void WiMeshNs3::GetNodesRemainingEnergyAtTime (double time, vector<double>* energy)
{
  Simulator::Schedule (Seconds (time), &WiMeshNs3::DoGetNodesRemainingEnergy, this, energy);
}


void WiMeshNs3::DoGetNodesRemainingEnergy (vector<double>* energy)
{
  energy->resize(ESnodes.GetN());

  for (uint32_t i = 0; i < ESnodes.GetN(); i++)
    energy->at(i) = ESnodes.Get(i)->GetRemainingEnergy();
}


void WiMeshNs3::ScheduleFailureEvent (FailureEvent& f, vector<int> &cumRadios, stringstream &fname)
{
  int _node;
  unsigned int _channel;
  vector<int>::iterator res;
  
  switch (f.type)
  {
    case 1:
      Simulator::Schedule (Seconds (f.time), &WiMeshNs3::DoGenerateNoise, this, f.node, f.channel, f.noise);
      fname << "_" << f.node << "_" << f.channel;
      break;
      
    case 2:
      res = find_if(cumRadios.begin(), cumRadios.end(), bind2nd(greater<int>(), f.radio));
      if (res == cumRadios.end())
      {
	cerr << "No radio found with the given index (" << f.radio << "). Noise is not added" << endl << endl;
	return;
      }
      
      _node = res - cumRadios.begin();
      _channel = topology->Node(_node)->Info()->Radio(f.radio-(*res-topology->Node(_node)->Info()->numAvailRadios()))->Channel();

      if (!_channel)
      {
	cout << "Warning: the radio is already turned off. Noise is not added" << endl << endl;
	return;
      }

      Simulator::Schedule (Seconds (f.time), &WiMeshNs3::DoGenerateNoise, this, _node, _channel, f.noise);
      fname << "_" << _node << "_" << _channel;
      break;
      
    case 3:
      for (unsigned int i = 0; i < topology->Node(f.node)->Info()->numAvailRadios(); i++)
	if ((_channel = topology->Node(f.node)->Info()->Radio(i)->Channel()))
	{
	  Simulator::Schedule (Seconds (f.time), &WiMeshNs3::DoGenerateNoise, this, f.node, _channel, f.noise);
	}
      fname << "_n" << f.node;
      break;
  }
}

WiMeshNs3::ChannelSwitch::ChannelSwitch (double time, int node, unsigned int fromCh, unsigned int toCh)
  : m_time (time),
    m_node (node),
    m_fromCh (fromCh),
    m_toCh (toCh)
{
}

void WiMeshNs3::ScheduleChannelSwitch (double time, int node, unsigned int fromCh, unsigned int toCh)
{
  ch_switches.push_back (ChannelSwitch (time, node, fromCh, toCh));
}


RadioIface*
WiMeshNs3::GetRadioPtr (Ptr<WifiNetDevice> wnd)
{
  int n = wnd->GetNode()->GetId();
  unsigned int channel = wnd->GetPhy()->GetChannelNumber() / channelsep;

  int ri = topology->Node(n)->Info()->findRadioWithChannel(channel);

  if (ri < 0)
    return NULL;

  return &(*topology->Node(n)->Info()->Radio(ri));
}


void WiMeshNs3::DoCreateNodes()
{
  // create the nodes
  meshnodes.Create (topology->GetNumNodes());
}


void WiMeshNs3::CreateNs3MeshNodes ()
{  
  DoCreateNodes ();

  // set the position of nodes
  MobilityHelper mobility;
  Ptr<ListPositionAllocator> positionAlloc = CreateObject<ListPositionAllocator> ();
  
  for (TGraphNodeIterator n=topology->NodeBegin(); n!=topology->NodeEnd(); n++)
    positionAlloc->Add (Vector (n->Info()->Xpos(), n->Info()->Ypos(), 0.0));

  mobility.SetPositionAllocator (positionAlloc);
  mobility.SetMobilityModel ("ns3::ConstantPositionMobilityModel");
  mobility.Install (meshnodes);
}

void WiMeshNs3::CreateNs3Hosts ()
{
  PointToPointHelper pointToPoint;
  InternetStackHelper internet;
  NetDeviceContainer hostdevices;
  Ipv4AddressHelper address;
  Ipv4StaticRoutingHelper ipv4RoutingHelper;
  ApplicationContainer apps;
  int64_t streamIndex = GetOnOffRandomVarStreamNumber (); // used to get consistent traffic load across scenarios
  Ptr<Ipv4StaticRouting> staticRouting;
  Ptr<Node> nodep;
  uint16_t port = 9000;
  string prefix("192.168.");
  
  pointToPoint.SetDeviceAttribute ("DataRate", StringValue ("100Mbps"));
  // set a large MTU so that the point to point link does not become a bottleneck
  pointToPoint.SetDeviceAttribute ("Mtu", UintegerValue (5000));
  pointToPoint.SetChannelAttribute ("Delay", StringValue ("1ms"));  

  // check that e2e flows are consistent and obtain the set of demands
  topology->checkE2Eflows(&demands);

  set<int> edges;
  
  // find all the edge nodes
  for (TDemands::iterator d = demands.begin (); d != demands.end (); d++)
  {
    edges.insert (d->first.first);
    edges.insert (d->first.second);
  }
  
  // associates each WiMesh node id with the index in the node container
  vector<uint32_t> index (topology->GetNumNodes());

  // create a host for each edge node and a point-to-point channel between them
  for (set<int>::iterator i = edges.begin (); i != edges.end (); i++)
  {
    index[*i] = hostnodes.GetN ();
    nodep = CreateObject<Node> ();
    hostnodes.Add (nodep);
    internet.Install (nodep);
    
    hostdevices = pointToPoint.Install (nodep, meshnodes.Get (*i));

    // assign IP addresses
    stringstream netaddr;
    netaddr << prefix << *i;
    address.SetBase ((netaddr.str() + ".0").c_str(), "255.255.255.0");
    address.Assign (hostdevices);
    
    // set default route on the host node
    staticRouting = ipv4RoutingHelper.GetStaticRouting (nodep->GetObject<Ipv4> ());
    staticRouting->SetDefaultRoute (Ipv4Address ((netaddr.str() + ".2").c_str ()), 1);
  }
    
  double mean_t_on = (t_on_min + t_on_max) / 2,
         mean_t_off = (t_off_min + t_off_max) / 2;
  double coeff = (mean_t_on + mean_t_off) / mean_t_on;
  int dem_id = 0;
    
  // install an OnOff application on the hosts
  for (TDemands::iterator d = demands.begin (); d != demands.end (); d++, dem_id++)
  {
    stringstream remote;
    remote << prefix << d->first.second << ".1";
    
    if (trafficType == "TraceBased")
    {
      UdpTraceBasedClientHelper udpTr (Ipv4Address (remote.str().c_str()), port, traceFiles[dem_id % traceFiles.size()]);
      
      apps = udpTr.Install (hostnodes.Get (index[d->first.first]));
    }
    else if (d->second) // the default is OnOff
    {
      OnOffHelper onOff(transportProto, InetSocketAddress (Ipv4Address (remote.str().c_str()), port));

      onOff.SetAttribute ("DataRate", DataRateValue (DataRate ((uint64_t)(d->second * 1000000 * coeff))));
      onOff.SetAttribute ("PacketSize", UintegerValue (packetsize));
//    For ns < 3.18 
//    onOff.SetAttribute ("OnTime", RandomVariableValue (UniformVariable (t_on_min, t_on_max)));
//    onOff.SetAttribute ("OffTime", RandomVariableValue (UniformVariable (t_off_min, t_off_max)));
     
      stringstream ontime, offtime;
      
      ontime << "ns3::UniformRandomVariable[Min=" << t_on_min << "|Max=" << t_on_max << "]";
      offtime << "ns3::UniformRandomVariable[Min=" << t_off_min << "|Max=" << t_off_max << "]";
      
      onOff.SetAttribute ("OnTime", StringValue (ontime.str()));
      onOff.SetAttribute ("OffTime", StringValue (offtime.str()));
      
      apps = onOff.Install (hostnodes.Get (index[d->first.first]));

      Ptr<OnOffApplication> ooa = DynamicCast<OnOffApplication> (apps.Get (0));
      if (ooa)
	streamIndex += ooa->AssignStreams (streamIndex);
    }

    apps.Start (Seconds (0.01));
    apps.Stop (Seconds (duration));

    // Create a packet sink to receive these packets
    PacketSinkHelper sink (transportProto, 
			  InetSocketAddress (Ipv4Address::GetAny (), port++));
    apps = sink.Install (hostnodes.Get (index[d->first.second]));
    apps.Start (Seconds (0.01));
  }

  if (pcapEnabled)
    pointToPoint.EnablePcapAll("host");
}

int WiMeshNs3::HostIPaddressToNodeId (const Ipv4Address& addr)
{
  uint8_t buf[4];

  addr.Serialize (buf);

  // the node id is stored in the 3rd byte

  return (int)(buf[2]);
}

bool WiMeshNs3::CheckDemand (int src, int dst, double f, bool strict)
{
  if (demands.find (pair<int,int> (src,dst)) == demands.end ())
    return false;
  
  if (strict && trunc(demands[pair<int,int> (src,dst)]) != trunc(f))
    return false;
  
  return true;
}

void WiMeshNs3::RunSimulation ()
{
  DoInit ();
  CreateNs3MeshNodes ();
  CreateNs3MeshDevices ();
  SetRSMoptions ();
  CreateNs3Hosts ();
  if (enableEnergyMeter)
    DoAttachEnergySources();

  // radios that are initially set to channel 0 are put in the sleep mode. This must be done after the Wifi
  // radio energy models are attached to them, otherwise the associated listeners are not notified that the
  // radios are put in the sleep mode (and the energy consumption is not computed correctly)
  for (uint32_t i = 0; i < radiosInitiallyOff.GetN (); i++)
    DynamicCast<WifiNetDevice>(radiosInitiallyOff.Get (i))->GetPhy ()->SetSleepMode ();

  ConfigureForwarding ();

  if (pcapEnabled)
    wifiPhy.EnablePcap ("mesh", meshdevices);

  Mac48Address ma("00:00:00:00:00:01");

  stringstream fname;

  fname << (transportProto == "ns3::UdpSocketFactory" ? "udp_" : "tcp_") << WMS.simulId() << "_";
  if (stack == "MplsSplitting")
    fname << "mpls";
  else if (stack == "ns3::LayerStack")
    fname << "layer";
  else if (stack ==  "ns3::FlameStack")
    fname << "flame";
  else if (stack == "ns3::Dot11sStack")
    fname << "dot11s";

  // find the vector of cumulative number of radios
  vector<int> cumRadios;
  int cum = 0;
  for (TGraphNodeIterator u = topology->NodeBegin(); u != topology->NodeEnd(); u++)
    cumRadios.push_back (cum += u->Info()->numAvailRadios());


  // schedules failures, if any
  for (vector<FailureEvent>::iterator e = failures.begin (); e != failures.end (); ++e)
  {
    ScheduleFailureEvent (*e, cumRadios, fname);
  }

  // schedules channel switches, if any
  for (vector<ChannelSwitch>::iterator cs = ch_switches.begin(); cs != ch_switches.end(); cs++)
    Simulator::Schedule (Seconds (cs->m_time), &WiMeshNs3::DoChannelSwitch, this, cs->m_node, cs->m_fromCh, cs->m_toCh);

  if (enableFlowMonitor)
    WMflowmon.Enable (measure_interval, duration, measure_perflow, string("flow_") + fname.str ());

  // force the simulator to terminate
  Simulator::Stop (Seconds(duration));
  Simulator::Run ();

  PrintConfiguration ();

  if (enableEnergyMeter)
    DoPrintEnergyInfo ((string("energy_")+fname.str()).c_str());

  Simulator::Destroy ();
}

void
WiMeshNs3::PrintObjConf (ostream& os)
{
  if (topology)
    os << "Topology name: " << topology->Toponame() << endl;
  os << "Channel separation: " << channelsep << endl << endl;

  if (trafficType == "TraceBased")
  {
    os << "Traffic type: " << trafficType << endl;
    for (vector<string>::iterator f = traceFiles.begin(); f != traceFiles.end(); f++)
      os << *f << endl;
  }
  else
  {
    os << "Traffic type: OnOff" << endl
       << "Ton ~ U(" << t_on_min << "," << t_on_max << ")     Toff ~ U(" << t_off_min << "," << t_off_max << ")" 
       << endl << endl << "Packet size: " << packetsize;
  }
  
  os << endl << "Transport protocol: " << transportProto << endl
     << "Duration: " << duration << endl << endl;
     
  for (vector<ChannelSwitch>::iterator cs = ch_switches.begin(); cs != ch_switches.end(); cs++)
    os << "Channel switch at time " << cs->m_time << " on node " << cs->m_node << " from channel " << cs->m_fromCh 
       << " to channel " << cs->m_toCh << endl;
}

void WiMeshNs3::DoPrintEnergyInfo (const char* fname)
{
  ofstream ofs (fname);
  vector<double> standbyTime;

  ofs << "# Per node energy consumption (node, radio1, radio2, ...)"  << endl << endl
      << "EnergyConsumption = {" << endl;

  // compute the energy consumed by each node and write results to file
  for (TGraphNodeIterator u = topology->NodeBegin(); u != topology->NodeEnd(); u++)
  {
    DeviceEnergyModelContainer demc = ESnodes.Get(u->Id())->FindDeviceEnergyModels ("ns3::WiMeshNodeEnergyModel");
    for (DeviceEnergyModelContainer::Iterator i = demc.Begin(); i != demc.End(); i++)
    {
      // required to update ns3::WiMeshNodeEnergyModel::m_totalEnergyConsumption
      (*i)->ChangeState (WiMeshNodeEnergyModel::STANDBY);
      ofs << " [ " << (*i)->GetTotalEnergyConsumption() << " ";
      standbyTime.push_back (DynamicCast<WiMeshNodeEnergyModel>(*i)->GetStandbyTime());
    }

    demc = ESnodes.Get(u->Id())->FindDeviceEnergyModels ("ns3::WifiRadioEnergyModel");
    for (DeviceEnergyModelContainer::Iterator i = demc.Begin(); i != demc.End(); i++)
    {
      // required to update ns3::WifiRadioEnergyModel::m_totalEnergyConsumption
      (*i)->ChangeState (WifiPhy::IDLE);
      ofs << (*i)->GetTotalEnergyConsumption() << " ";
    }

    ofs << "]" << endl;
  }
  ofs << "};" << endl << endl;

  ofs << "# Time spent by each node in standby mode (seconds)" << endl << endl
      << "StandbyTime = [" << endl;
  copy (standbyTime.begin(), standbyTime.end(), ostream_iterator<double>(ofs,"\n"));
  ofs << "];" << endl;

  double totalStandbyTime = 0.0, totalStandbyTimeExcludingNodesOff = 0.0;
  
  for (int u = 0; u < topology->GetNumNodes(); u++)
  {
    totalStandbyTime += standbyTime[u];
    if (topology->Node(u)->Info()->countRadiosWithChannel(0) != topology->Node(u)->Info()->GetNumRadios())
      totalStandbyTimeExcludingNodesOff += standbyTime[u];
  }

  ofs << endl << "TotalStandbyTime = " << totalStandbyTime << ";" << endl;
  ofs << endl << "TotalStandbyTimeExcludingNodesOff = " << totalStandbyTimeExcludingNodesOff << ";" << endl;

  ofs.close();
}

void
WiMeshNs3::GetTxRatePower (int myId, Mac48Address neighbor, uint32_t &rate, uint32_t &power)
{
  int channel;
  int neighbId = DoFindNeighborByMacAddress (myId, neighbor, channel);
  
  // if the link is not present in topology, return the basic rate
  if (myId < 0 || myId >= (int)topology->GetNumNodes() || neighbId < 0 || neighbId >= (int)topology->GetNumNodes() )
  {
    rate = 0;
    power = WMS.defaultTxPowerLevel();
    return;
  }
  
  TGraphNeighborIterator n = topology->Node(myId)->FindNeighbor (neighbId, channel);

  if (n == topology->Node(myId)->NeighborEnd ()) 
  {
    rate = 0;
    power = WMS.defaultTxPowerLevel();
    return;
  }

  rate = n->Rate();
  power = n->TxPower();
}

void
WiMeshNs3::ReceiveFinalDataFailed (const WifiRemoteStationManager* me, Mac48Address neighbor)
{
  int myId = DoFindNodeByStationManager (me);
  DoReceiveFinalDataFailed (myId, neighbor);
}


int
WiMeshNs3::DoFindNodeByStationManager (const WifiRemoteStationManager* me)
{
  Ptr<Node> n;
  bool found = false;
  
  for (NetDeviceContainer::Iterator i = meshdevices.Begin (); i != meshdevices.End () && !found; ++i)
    if (GetPointer<WifiRemoteStationManager>((*i)->GetObject<WifiNetDevice> ()->GetRemoteStationManager ()) == me)
    {
      found = true;
      n = (*i)->GetNode ();
    }
    
  assert (found);
  
  for (uint32_t i = 0; i < meshnodes.GetN (); ++i)
    if (meshnodes.Get(i)->GetId () == n->GetId ())
      return i;
  
  // should never get here
  return topology->GetNumNodes();
}

int
WiMeshNs3::DoFindNeighborByMacAddress (int me, Mac48Address neighbor, int& channel, Ptr<WifiNetDevice>* device)
{
  Ptr<WifiNetDevice> wnd;

  // for each neighbor of mine
  for (TGraphNeighborIterator n = topology->Node(me)->NeighborBegin (); n != topology->Node(me)->NeighborEnd (); n++)
  {
    // check the address of all of its Wifi interfaces
    for (uint32_t i = 0; i < meshnodes.Get (n->Id())->GetNDevices (); ++i)
    {
      wnd = meshnodes.Get (n->Id())->GetDevice (i)->GetObject<WifiNetDevice> ();
      // if it is a WifiNetDevice and it has neighbor as address
      if (wnd != 0 && Mac48Address::ConvertFrom (wnd->GetAddress ()) == neighbor)
      {
	channel = wnd->GetPhy ()->GetChannelNumber () / channelsep;
	if (device)
	  *device = wnd;
	return n->Id();
      }
    }
  }
  // we might get here when not using Mpls or Layer
  return topology->GetNumNodes();
}

void
WiMeshNs3::DoGenerateNoise (int node, unsigned int channel, double noise)
{
  Ptr<WifiNetDevice> wnd;
  
  for (uint32_t i = 0; i < meshnodes.Get (node)->GetNDevices (); ++i)
  {
    wnd = meshnodes.Get (node)->GetDevice (i)->GetObject<WifiNetDevice> ();
    // if it is a WifiNetDevice and it is set to the given channel
    if (wnd != 0 && wnd->GetPhy ()->GetChannelNumber () / channelsep == channel)
    {
      cerr << "Adding a " << noise << "db noise around the radio on node " << node << " using channel " << channel << endl;
      wnd->GetPhy ()->GetObject<YansWifiPhy> ()->SetRxNoiseFigure (1e10);
    }
  }
}

void
WiMeshNs3::DoChannelSwitch (int node, unsigned int fromCh, unsigned int toCh)
{
  Ptr<WifiNetDevice> wnd;
  
  for (uint32_t i = 0; i < meshnodes.Get (node)->GetNDevices (); ++i)
  {
    wnd = meshnodes.Get (node)->GetDevice (i)->GetObject<WifiNetDevice> ();
    // if it is a WifiNetDevice and it is set to the given channel
    if (wnd != 0 && wnd->GetPhy ()->GetChannelNumber () / channelsep == fromCh)
    {
      cerr << "Switching a radio on node " << node << " from channel " << fromCh << " to channel ";
      wnd->GetPhy ()->SetChannelNumber (toCh * channelsep);
      cerr << wnd->GetPhy ()->GetChannelNumber () / channelsep  << endl;
      break;
    }
  }
}


WiMeshNs3Ptr::WiMeshNs3Ptr ()
  : m_ptr (0)
{
}

WiMeshNs3Ptr::WiMeshNs3Ptr (WiMeshNs3* ptr)
  : m_ptr (ptr)
{
}

WiMeshNs3*
WiMeshNs3Ptr::GetPtr ()
{
  return m_ptr;
}


ostream&
operator<< (ostream& os, const WiMeshNs3Ptr& p)
{
  os << "this does not make sense";
  return os;
}

istream&
operator>>(istream& is, WiMeshNs3Ptr& p)
{
  return is;
}


ATTRIBUTE_HELPER_CPP (WiMeshNs3Ptr);

