# find ns-3
#
# exports:
#
#   NS3_FOUND
#   NS3_INCLUDE_DIRS
#   NS3_LIBRARIES
#

include(FindPkgConfig)
include(FindPackageHandleStandardArgs)

# Use pkg-config to get hints about paths
# NOTE: On most systems, pkg-config looks in /usr/lib/pkgconfig, /usr/share/pkgconfig, /usr/local/lib/pkgconfig
#       and /usr/local/share/pkgconfig for .pc files. It will additionally look in the colon-separated (on Windows,
#       semicolon-separated) list of directories specified by the PKG_CONFIG_PATH environment variable.

# Since the version number is embedded into the ns-3 library names, we cannot use
# pkg-config directly to find them

# find the default path used by pkg-config
execute_process (COMMAND ${PKG_CONFIG_EXECUTABLE} --variable pc_path pkg-config
                 OUTPUT_VARIABLE PkgConfPath)

# remove trailing newlines
string (REGEX REPLACE "(\r?\n)+$" "" PkgConfPath  "${PkgConfPath}")

# if the PKG_CONFIG_PATH environment variable is defined, append its content
if (DEFINED ENV{PKG_CONFIG_PATH})
  set (PkgConfPath "${PkgConfPath}:$ENV{PKG_CONFIG_PATH}")
endif ()

# replace ':' with ';' so that it is seen as a list of directories
string (REPLACE ":" ";" PkgConfPath "${PkgConfPath}")

foreach (dir ${PkgConfPath})

  # check if there is any version of the ns-3 core library in this directory
  file (GLOB fname RELATIVE ${dir} "${dir}/libns3*-core-*pc")

  # for each file that has been found (if any)
  foreach (fi ${fname})

    if (NOT "${fi}" STREQUAL "")

      # determine version
      string (FIND ${fi} "dev" pos)

      if (NOT pos EQUAL -1)
        set (Ns3Version "3-dev")
      else ()
        string (FIND ${fi} "-" pos)
        math (EXPR len "${pos} - 5")
        string (SUBSTRING ${fi} 5 ${len} Ns3Version)
      endif ()

      # determine build type (debug or optimized)
      string (FIND ${fi} "-" pos REVERSE)
      string (FIND ${fi} "." posdot REVERSE)
      math (EXPR pos "${pos} + 1")
      math (EXPR len "${posdot} - ${pos}")
      string (SUBSTRING ${fi} ${pos} ${len} BuildType)

    endif()

  endforeach()

endforeach()

# the layer2.5 component is dealt with in a special way

list (FIND Ns3_FIND_COMPONENTS "layer2.5" l2.5_index)

# if layer2.5 has been specified by the caller, remove it from Ns3_FIND_COMPONENTS.
# Further actions are taken afterwards

if (NOT l2.5_index EQUAL -1)
  
  list (REMOVE_AT Ns3_FIND_COMPONENTS ${l2.5_index})

  # check if the mesh component has been included in the current components 
  list (FIND Ns3_FIND_COMPONENTS "mesh" mesh_index)
  
  # if it has not been included, append it to the list of components
  if (mesh_index EQUAL -1)
    list (APPEND Ns3_FIND_COMPONENTS "mesh")
    set (Ns3_FIND_REQUIRED_mesh ${Ns3_FIND_REQUIRED_layer2.5})
  endif()
  
endif()


# process the list of components

foreach (component ${Ns3_FIND_COMPONENTS})

  set (compname "libns${Ns3Version}-${component}-${BuildType}")
  
  string (TOUPPER ${component} upper_component)
  
  if (Ns3_FIND_REQUIRED_${component})
    pkg_check_modules (NS3_${upper_component}_PKGCONF REQUIRED ${compname})
  else()
    pkg_check_modules (NS3_${upper_component}_PKGCONF ${compname})
  endif()

  # Include dir
  find_path(Ns3_${upper_component}_INCLUDE_DIR
    NAMES ns3/${component}-module.h
    PATHS ${NS3_${upper_component}_PKGCONF_INCLUDE_DIRS}
  )

  set (libname ns${NS3_${upper_component}_PKGCONF_VERSION}-${component}-${BuildType})

  # Library
  find_library(Ns3_${upper_component}_LIBRARY
    NAMES ${libname}
    PATHS ${NS3_${upper_component}_PKGCONF_LIBRARY_DIRS}
  )

  if (Ns3_${upper_component}_INCLUDE_DIR AND Ns3_${upper_component}_LIBRARY)
    list (APPEND __Ns3_INCLUDE_DIRS ${Ns3_${upper_component}_INCLUDE_DIR})
    list (APPEND __Ns3_LIBRARIES ${Ns3_${upper_component}_LIBRARY})
  endif()
  
  FIND_PACKAGE_HANDLE_STANDARD_ARGS(Ns3_${component}
                                    FOUND_VAR
                                      Ns3_${component}_FOUND
                                    REQUIRED_VARS
                                      Ns3_${upper_component}_LIBRARY
                                      Ns3_${upper_component}_INCLUDE_DIR)

endforeach()

if (__Ns3_INCLUDE_DIRS)
  list(REMOVE_DUPLICATES __Ns3_INCLUDE_DIRS)
endif()

set (Ns3_INCLUDE_DIRS ${__Ns3_INCLUDE_DIRS} CACHE INTERNAL "ns-3 include dirs")
set (Ns3_LIBRARIES ${__Ns3_LIBRARIES} CACHE INTERNAL "ns-3 libraries")

FIND_PACKAGE_HANDLE_STANDARD_ARGS(Ns3
                                  FOUND_VAR
                                    Ns3_FOUND
                                  REQUIRED_VARS
                                    Ns3_INCLUDE_DIRS
                                    Ns3_LIBRARIES
                                  HANDLE_COMPONENTS)  


# Continue dealing with layer2.5, if present

if (NOT l2.5_index EQUAL -1)

  if (NOT Ns3_mesh_FOUND)

    # layer2.5 is not installed either
    unset(Ns3_layer2.5_FOUND)

    if (Ns3_FIND_REQUIRED_layer2.5)
      message (FATAL_ERROR "Could not find NS3_LAYER2.5")
    else()
      message (WARNING "Could not find NS3_LAYER2.5")
    endif()
  
  else()
  
    # the mesh component has been found, search for layer-protocol.h
    find_path(Ns3_LAYER2.5_INCLUDE_DIR
      NAMES ns3/layer-protocol.h
      PATHS ${NS3_MESH_PKGCONF_INCLUDE_DIRS}
    )
    
    FIND_PACKAGE_HANDLE_STANDARD_ARGS(Ns3_layer2.5
                                      FOUND_VAR
                                        Ns3_layer2.5_FOUND
                                      REQUIRED_VARS
                                        Ns3_LAYER2.5_INCLUDE_DIR)

  endif()

endif()
